﻿Imports System.Drawing

Module CurveFunctions
    Private Const TINY As Double = 0.0000001
    Private Const G_A_MIN As Double = 300
    Private Const G_A_MAX As Double = 400
    Private Const G_B_MIN As Double = 200
    Private Const G_B_MAX As Double = 300
    Private Const G_C_MIN As Double = 0
    Private Const G_C_MAX As Double = 10
    Private Const G_D_MIN As Double = 0
    Private Const G_D_MAX As Double = 5

    ' Return the function for x, a, b, and c.
    Public Function F(ByVal x As Double, ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double) As Double
        Return a - b * Math.Exp(-c * (x ^ d))
    End Function

    ' dF/da
    Private Function dFda(ByVal x As Double, ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double) As Double
        Return 1
    End Function

    ' dF/db
    Private Function dFdb(ByVal x As Double, ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double) As Double
        Return -Math.Exp(-c * (x ^ d))
    End Function

    ' dF/dc
    Private Function dFdc(ByVal x As Double, ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double) As Double
        Return -b * Math.Exp(-c * (x ^ d)) * (-(x ^ d))
    End Function

    ' dF/dd
    Private Function dFdd(ByVal x As Double, ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double) As Double
        Return -b * Math.Exp(-c * (x ^ d)) * (-c * (x ^ d) * Math.Log(x))
    End Function

    ' Return the sum of errors squared for the given points and parameters.
    Public Function ErrorSquared(ByVal pts As List(Of PointF), ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double) As Double
        Dim total_error As Double = 0
        For i As Integer = 0 To pts.Count - 1
            Dim new_error As Double = pts(i).Y - F(pts(i).X, a, b, c, d)
            If (total_error > 1.0E+15) Then Exit For
            total_error += new_error * new_error
        Next i

        Return total_error
    End Function

    ' dE/da
    Private Function dEda(ByVal pts As List(Of PointF), ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double) As Double
        Dim total As Double = 0
        For i As Integer = 0 To pts.Count - 1
            total += 2 * (F(pts(i).X, a, b, c, d) - pts(i).Y) * _
                dFda(pts(i).X, a, b, c, d)
        Next i
        Return total
    End Function

    ' dE/db
    Private Function dEdb(ByVal pts As List(Of PointF), ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double) As Double
        Dim total As Double = 0
        For i As Integer = 0 To pts.Count - 1
            total += 2 * (F(pts(i).X, a, b, c, d) - pts(i).Y) * _
                dFdb(pts(i).X, a, b, c, d)
        Next i
        Return total
    End Function

    ' dE/dc
    Private Function dEdc(ByVal pts As List(Of PointF), ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double) As Double
        Dim total As Double = 0
        For i As Integer = 0 To pts.Count - 1
            total += 2 * (F(pts(i).X, a, b, c, d) - pts(i).Y) * _
                dFdc(pts(i).X, a, b, c, d)
        Next i
        Return total
    End Function

    ' dE/dd
    Private Function dEdd(ByVal pts As List(Of PointF), ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double) As Double
        Dim total As Double = 0
        For i As Integer = 0 To pts.Count - 1
            total += 2 * (F(pts(i).X, a, b, c, d) - pts(i).Y) * _
                dFdd(pts(i).X, a, b, c, d)
        Next i
        Return total
    End Function

    ' Find a good curve fit.
    Public Function FindGoodFit(ByVal pts As List(Of PointF), _
     ByRef best_a As Double, ByRef best_b As Double, ByRef best_c As Double, ByRef best_d As Double, _
     ByVal num_steps As Integer, ByVal max_iterations As Integer) As Double
        best_a = 0
        best_b = 0
        best_c = 0
        best_d = 0
        Dim test_a As Double
        Dim test_b As Double
        Dim test_c As Double
        Dim test_d As Double

        Dim g_a_step As Double = (G_A_MAX - G_A_MIN) / (num_steps - 1)
        Dim g_b_step As Double = (G_B_MAX - G_B_MIN) / (num_steps - 1)
        Dim g_c_step As Double = (G_C_MAX - G_C_MIN) / (num_steps - 1)
        Dim g_d_step As Double = (G_D_MAX - G_D_MIN) / (num_steps - 1)

        Dim best_error2 As Double = Double.MaxValue
        For a As Double = G_A_MIN To G_A_MAX Step g_a_step
            For b As Double = G_B_MIN To G_B_MAX Step g_b_step
                For c As Double = G_C_MIN To G_C_MAX Step g_c_step
                    For d As Double = G_D_MIN To G_D_MAX Step g_d_step
                        Dim test_error2 As Double = FollowGradient( _
                            pts, a, b, c, d, _
                            test_a, test_b, test_c, test_d, max_iterations)
                        If (test_error2 < best_error2) Then
                            best_a = test_a
                            best_b = test_b
                            best_c = test_c
                            best_d = test_d
                            best_error2 = test_error2
                        End If
                    Next d
                Next c
            Next b
        Next a

        Return Math.Sqrt(best_error2)
    End Function

    ' Starting at (X, Y), follow the error
    ' function's gradient to try to improve it.
    Private Function FollowGradient(ByVal pts As List(Of PointF), _
     ByVal a As Double, ByVal b As Double, ByVal c As Double, ByVal d As Double, _
     ByRef best_a As Double, ByRef best_b As Double, ByRef best_c As Double, ByRef best_d As Double, _
     ByVal max_iterations As Integer) As Double
        Const CUTOFF_ERROR As Double = 1.0

        ' Set the initial distance to move.
        Dim dist As Double = 0.25

        ' Start with the initial point.
        best_a = a
        best_b = b
        best_c = c
        best_d = d
        Dim best_error2 As Double = ErrorSquared(pts, a, b, c, d)

        For iteration As Integer = 1 To max_iterations
            ' Get the gradient at this point.
            Dim na As Double = dEda(pts, a, b, c, d)
            Dim nb As Double = dEdb(pts, a, b, c, d)
            Dim nc As Double = dEdc(pts, a, b, c, d)
            Dim nd As Double = dEdd(pts, a, b, c, d)

            ' Normalize it.
            Dim length As Double = na * na + nb * nb + nc * nc + nd * nd
            length = Math.Sqrt(length)
            ' If the length is too small, return the result so far.
            If (length < TINY) Then Return best_error2
            na = -na / length
            nb = -nb / length
            nc = -nc / length
            nd = -nd / length

            ' Try moving along the gradient.
            Dim va As Double = na * dist
            Dim vb As Double = nb * dist
            Dim vc As Double = nc * dist
            Dim vd As Double = nd * dist
            Dim test_error2 As Double = ErrorSquared(pts, a + va, b + vb, c + vc, d + vd)

            ' Try increasingly smaller vectors
            ' until we find an improvement.
            While (test_error2 > best_error2)
                dist /= 2
                If (dist < TINY) Then
                    ' We're not moving far enough. Stop.
                    ' Return the result so far.
                    Return best_error2
                End If
                va = na * dist
                vb = nb * dist
                vc = nc * dist
                vd = nd * dist
                test_error2 = ErrorSquared(pts, a + va, b + vb, c + vc, d + vd)
            End While

            ' Hopefully at this point we have found an improvement.
            a += va
            b += vb
            c += vc
            d += vd
            best_a = a
            best_b = b
            best_c = c
            best_d = d
            best_error2 = test_error2

            ' If the error is small enough, return.
            If (best_error2 < CUTOFF_ERROR) Then Return best_error2
        Next iteration

        ' Return the error.
        Return best_error2
    End Function
End Module
